#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Assignment-Free REDOR Analysis

Written by Noah H. Somberg and Mei Hong

Correspondence to: meihong@mit.edu
"""
import numpy as np
import structure_tools
import matplotlib.pyplot as plt
import matplotlib
from scipy.spatial.transform import Rotation as R
import pandas as pd

# This makes the text on the figures editable when it's output
matplotlib.rcParams['pdf.fonttype'] = 42
matplotlib.rcParams['ps.fonttype'] = 42

# Set the default sans-serif font to Arial
matplotlib.rcParams['font.sans-serif'] = "Arial"
matplotlib.rcParams['font.family'] = "sans-serif"
matplotlib.rcParams['font.size'] = 12

def get_coupling_1A():
    '''
    Get the 1 Angstrom CF coupling, including scaling

    '''
    COUPLING_SCALE = 0.48
    coupling_1A = 2*np.pi*120120*0.941286*0.251504*COUPLING_SCALE
    return coupling_1A

def print_pymol_commands(coord_list,selection_name='fit_atoms',atom_name='FP'):
    '''
    Given a list of coords, outputs a set of pymol commands to create pseudoatoms to the terminal

    '''
    i = 0
    
    if isinstance(coord_list, np.ndarray):
        coord_list = np.vstack(coord_list)
    
    for coord in coord_list:
        x = coord[0]
        y = coord[1]
        z = coord[2]
        xyz = '['+str(x)+','+str(y)+','+str(z)+']'
        output = 'pseudoatom tmp'+str(i)+',name='+atom_name+',pos='+xyz
        i += 1
        print(output)
        
    print('select '+selection_name+', name '+atom_name)
    
def write_pymol_to_file(coord_list,file_name,selection_name='fit_atoms',atom_name='FP'):
    '''
    Given a list of coords, outputs a set of pymol commands to create pseudoatoms to a file.

    '''
    i = 0
    
    if isinstance(coord_list, np.ndarray):
        coord_list = np.vstack(coord_list)
    
    with open(file_name,'w') as f:
        for coord in coord_list:
            x = coord[0]
            y = coord[1]
            z = coord[2]
            xyz = '['+str(x)+','+str(y)+','+str(z)+']'
            output = 'pseudoatom tmp'+str(i)+',name='+atom_name+',pos='+xyz
            i += 1
            f.write(output+'\n')
        f.write('select '+selection_name+', name '+atom_name)
        
def write_pymol_to_file_rmsds(coord_list,rmsds,file_name,selection_name='fit_atoms',atom_name='FP'):
    '''
    Given a list of coords and RMSD vals, writes them to a pymol file categorized by the RMSD

    '''
    
    i = 0
    
    if isinstance(coord_list, np.ndarray):
        coord_list = np.vstack(coord_list)
    
    with open(file_name,'w') as f:
        for j in range(len(coord_list)):
            coord = coord_list[j]
            rmsd = rmsds[j]
            
            if rmsd < 0.02:
                atom_name = 'FA'
            elif rmsd < 0.03:
                atom_name = 'FB'
            elif rmsd < 0.04:
                atom_name = 'FC'
            else:
                atom_name = 'FD'
            
            x = coord[0]
            y = coord[1]
            z = coord[2]
            xyz = '['+str(x)+','+str(y)+','+str(z)+']'
            output = 'pseudoatom tmp'+str(i)+',name='+atom_name+',pos='+xyz
            i += 1
            f.write(output+'\n')
        f.write('select '+selection_name+', name '+atom_name)
                

def coords_to_distances(dephased_coordinates,dephasor_coordinates):
    """
    Takes in an N x 3 np.array representing atomic coordinates of the sites dephased and an M x 3 np.array representing the atomic coordinates of the dephasors.
    Returns an N x M np.array where each row is the distances from that dephased spin to each dephasor
    
    Parameters
    ----------
    dephased_coordinates : np.array (N x 3)
        Nx3 array of atomic coordinates to be dephased
    dephasor_coordinates : np.array (M x 3)
        Mx3 array of atomic coordinates of dephasors
        
    Returns
    -------
    dists : np.array (N x M)
        NxM array of distances from dephased spin N to dephasor spin M
    """
    
    recenter = dephased_coordinates[:, np.newaxis, :] - dephasor_coordinates[np.newaxis, :, :] # Recenter the coordinates
    dists = np.linalg.norm(recenter,axis=2) # compute the norm
    return dists
    
def second_moment_redor(distance_array,coup_1A,sim_time_axis):
    '''
    Takes in an (N x M) np.array of distances, where N is the number of residues atoms in the category, and M is the number of dephasors

    Parameters
    ----------
    distance_array : np.array (NxM)
        Array of distances from dephased spin N to dephasor spin M
    coup_1A : float
        1 Anstrom C-F coupling in RADIANS
    sim_time_axis : np.array
        dephasing time points in SECONDS

    Returns
    -------
    avg_curve : np.array
        the average redor curve, 1xlen(sim_time_axis)

    '''
    
    # Example distance array
    # 3 dephased spins, 2 dephasors
    # np.array([[5,10],[15,21],[7,3]])
    
    M2 = np.sum((coup_1A/(distance_array**3))**2,axis=1)/5
    all_curves = np.exp((-4*M2[:,np.newaxis] * (sim_time_axis**2))/(3*np.pi**2))
    avg_curve = np.sum(all_curves,axis=0)/len(all_curves)

    return avg_curve

def is_coordinate_valid(coordinate,bounding_box):
    '''
    Checks if coordinate is within bounding box

    Parameters
    ----------
    coordinate : np.array (1x3)
        [x,y,z]
    bounding_box : np.array (3x2)
        first row is [x_min, x_max], second is [y_min,y_max], etc

    Returns
    -------
    bool
        True if in the box, False otherwise.

    '''
    
    if (bounding_box[0,0] < coordinate[0] < bounding_box[0,1] and
        bounding_box[1,0] < coordinate[1] < bounding_box[1,1] and
        bounding_box[2,0] < coordinate[2] < bounding_box[2,1]):
        
        return True
    else:
        return False
        
        
def get_random_coordinate(bounding_box):
    '''
    Get a random coordinate within the bounding box

    Parameters
    ----------
    bounding_box : np.array (3x2)
        first row is [x_min, x_max], second is [y_min,y_max], etc

    Returns
    -------
    coordinate : np.array (1x3)
        [x,y,z] coordinate

    '''
    x = np.random.uniform(bounding_box[0,0],bounding_box[0,1])
    y = np.random.uniform(bounding_box[1,0],bounding_box[1,1])
    z = np.random.uniform(bounding_box[2,0],bounding_box[2,1])
    coordinate = np.array([x,y,z])
    return coordinate

def generate_tetrahedron(d_center_vertex):
    '''
    Given the distance from the center to a vertex, generate the coordinates of a tetrahedron

    '''
    
    # The distance from the center to the vertex is the circumradius
    rad = d_center_vertex

    # Side legnth
    s = (4 / np.sqrt(6)) * rad

    # Tetrahedron with side len 2
    v0 = np.array([1, 0, -1/np.sqrt(2)])
    v1 = np.array([-1, 0, -1/np.sqrt(2)])
    v2 = np.array([0, 1, 1/np.sqrt(2)])
    v3 = np.array([0, -1, 1/np.sqrt(2)])

    # Scale
    vertices = [tuple(s/2 * v) for v in [v0, v1, v2, v3]]

    return np.array(vertices)

def get_initial_angles():
    '''
    Get an intiial rotation for the drug

    Returns
    -------
    quat : Scipy.Rotation
        the rotation as a scipy rotation object.

    '''
    
    # Generate a random unit vector (axis of rotation)
    axis = np.random.randn(3)
    axis /= np.linalg.norm(axis)
    
    # Generate a random rotation angle within the allowed range
    angle = np.random.uniform(0, 360)
    
    # Create a rotation quaternion
    quat = R.from_rotvec(axis * angle)
    
    # Convert to Euler angles (roll, pitch, yaw)
    #roll, pitch, yaw = quat.as_euler('xyz', degrees=True)
    
    return quat
    
def get_new_angles(old_angles,max_angle_degrees):
    '''
    Given the old rotation, and maximum angle step, generate the new anlges

    Parameters
    ----------
    old_angles : Scipy.Rotation
        old rotation, as a scipy rotation onkect
    max_angle_degrees : int
        maximum change in angles in degrees

    Returns
    -------
    new_rot : Scipy.Rotation
        new rotation, as a scipy rotation onkect

    '''

    max_angle_radians = np.radians(max_angle_degrees)
    
    # Generate a random unit vector (axis of rotation)
    axis = np.random.randn(3)
    axis /= np.linalg.norm(axis)
    
    # Generate a random rotation angle within the allowed range
    angle = np.random.uniform(0, max_angle_radians)
    
    # Create a rotation quaternion
    quat = R.from_rotvec(axis * angle)
    
    new_rot = old_angles * quat
    
    return new_rot

def rotate_translate(vertices, new_center, new_angles):
    '''
    Rotates and translates the coordinates

    '''
    
    # Rotate
    vertices = new_angles.apply(vertices)
    
    # Translate 
    vertices = vertices+new_center
    
    return vertices

def get_experimental_data_EmrE():
    '''
    Returns the experimental data organized something like this:
    expt_data['ACA'] = {'res_types': 'DL',
                         'SS0': [0.96, 0.94, 0.89],
                         'err': None,
                         'atoms' : ['CA']
                         } 
    '''
    
    expt_time = np.array([128, 250, 380])/25000
    
    df = pd.read_csv('EMRE_integrals_CA_methyl_3.csv') # No cys
    
    integrals = {}
    sinos = {}
    
    for index, row in df.iterrows():
        
        type_key = row['TYPE_FUZZY']+'_'+row['ATOMS']
        
        if type_key in integrals.keys():
            integrals[type_key] += np.array(row[1:7])
        else:
            integrals[type_key] = np.array(row[1:7])
            sinos[type_key] = row['SINO']
            
    expt_data = {}
    for type_key in integrals.keys():
        
        single_data = {}
        
        single_data['res_types'] = type_key.split('_')[0]
        ss0 = integrals[type_key][1::2]/integrals[type_key][0::2]
        single_data['SS0'] = ss0
        
        S0_sinos = sinos[type_key]*(integrals[type_key][0::2]/integrals[type_key][0])
        
        single_data['err'] = ss0*((1/S0_sinos)**2 + (1/(S0_sinos*ss0))**2)**0.5
        
        single_data['atoms'] = type_key.split('_')[1].split()
        
        expt_data[type_key] = single_data
        
    return (expt_time,expt_data)

def get_experimental_data_GB1():
    '''
    Returns the experimental data organized something like this:
    expt_data['ACA'] = {'res_types': 'DL',
                         'SS0': [0.96, 0.94, 0.89],
                         'err': None,
                         'atoms' : ['CA']
                         } 
    '''
    
    expt_time = np.array([128, 250, 380])/25000
    
    df = pd.read_csv('GB1_integrals_forpy.csv')
    
    integrals = {}
    all_sinos = {}
    
    for index, row in df.iterrows():
        
        type_key = row['TYPE_FUZZY']+'_'+row['ATOMS']
        
        if type_key in integrals.keys():
            integrals[type_key] += np.array(row[1:7])
            joined = np.append(all_sinos[type_key],np.array(row[9:15]))
            all_sinos[type_key] = joined
        else:
            integrals[type_key] = np.array(row[1:7])
            all_sinos[type_key] = np.array(row[9:15])
            
    expt_data = {}
    for type_key in integrals.keys():
        
        single_data = {}
        
        single_data['res_types'] = type_key.split('_')[0]
        single_data['SS0'] = integrals[type_key][1::2]/integrals[type_key][0::2]
        
        #base_sino = np.sum((1/((all_sinos[type_key])[all_sinos[type_key] > 0]))**2)**-0.5
        
        base_sino = np.sum(all_sinos[type_key]**2)**0.5
        
        S0_sinos = (integrals[type_key][0::2] / integrals[type_key][0])*base_sino
        single_data['err'] = single_data['SS0']*((1/S0_sinos)**2 + (1/(S0_sinos*single_data['SS0']))**2)**0.5
        
        single_data['atoms'] = type_key.split('_')[1].split()
        
        expt_data[type_key] = single_data
        
    return (expt_time,expt_data)

def get_point_rmsd(dephasor_coordinates,coordinate_sets,expt_time,expt_data,coup_1A):
    '''
    For a set of dephasor coordinates, simulates REDOR and gets the RMSD to the expt. data

    Parameters
    ----------
    dephasor_coordinates : np.array
        numpy array of dephasor coordinates
    coordinate_sets : dict of np.arrays
        the coordinate sets for each residue category
    expt_time : np.array
        expt times in SECONDS
    expt_data : dict of np.arrays
        S/S0 data for each residue category
    coup_1A : float
        1 Angstrom C-F coupling, including scaling

    Returns
    -------
    rmsd : float
        RMSD for those coordinates

    '''
    
    
    all_dev = np.array([])
    for atom_type_key in expt_data.keys():
        
        entry = expt_data[atom_type_key]
        ss0_data = entry['SS0']
        
        dephased_coordinates = coordinate_sets[atom_type_key] # Get the atomic coordinates
        distances = coords_to_distances(dephased_coordinates,dephasor_coordinates) # Convert to distances

        redor_sim = second_moment_redor(distances,coup_1A,expt_time)

        all_dev = np.append(all_dev,((ss0_data - redor_sim)**2))
        
    rmsd = (all_dev.mean())**0.5
    return rmsd

def get_new_center(old_center,max_step,bounding_box):
    '''
    Get the new center without leaving the box
    
    Parameters
    ----------
    old_center : np.array (1x3)
        Old ligand center.
    max_step : float
        Maximum step the ligand can take.
    bounding_box : np.array (3x2)
        first row is [x_min, x_max], second is [y_min,y_max], etc

    Returns
    -------
    new_center : np.array (1x3)
       New ligand center.

    '''
    
    phi = np.random.uniform(0, 2 * np.pi)  # Uniform azimuthal angle
    costheta = np.random.uniform(-1, 1)   # Uniform cosine of polar angle
    theta = np.arccos(costheta)           # Convert to polar angle

    x = np.sin(theta) * np.cos(phi)
    y = np.sin(theta) * np.sin(phi)
    z = np.cos(theta)
    
    step = np.array([x,y,z])*np.random.uniform(0,max_step)
    new_center = old_center+step
    
    while not is_coordinate_valid(new_center, bounding_box):
        phi = np.random.uniform(0, 2 * np.pi)  # Uniform azimuthal angle
        costheta = np.random.uniform(-1, 1)   # Uniform cosine of polar angle
        theta = np.arccos(costheta)           # Convert to polar angle

        x = np.sin(theta) * np.cos(phi)
        y = np.sin(theta) * np.sin(phi)
        z = np.cos(theta)
        
        step = np.array([x,y,z])*np.random.uniform(0,max_step)
        new_center = old_center+step
        
    return new_center

def acceptanceFunction(T,deltaE):
    '''
    Acceptance function for simulated annealing

    Parameters
    ----------
    T : float
        Temperature.
    deltaE : float
        Energy difference.

    Returns
    -------
    bool

    '''
    
    if deltaE < 0:
        return True
    else:
        r = np.random.random()
        if r < np.exp(-deltaE / T):
            return True
        else:
            return False
        
def plot_one_redor_point(dephasor_coordinates,expt_time,expt_data,coord_sets,export=False):
    '''
    Plots the type-REDOR curves for a single dephasor coordinate

    '''
    
    coup_1A = get_coupling_1A()
    
    n_types = len(expt_data.keys())
    fig,axs = plt.subplots(1,n_types,figsize=(8, 6))
    
    sim_time = np.linspace(0,25,200)/1000
    
    i = 0
    for res_type_sets in expt_data.keys():
        
        entry = expt_data[res_type_sets]
        ss0_data = entry['SS0']
        
        dephased_coordinates = coord_sets[res_type_sets]
        
        distances = coords_to_distances(dephased_coordinates,dephasor_coordinates) # Convert to distances
        redor_sim = second_moment_redor(distances,coup_1A,sim_time)
        
        axs[i].plot(sim_time*1000,redor_sim)
        
        axs[i].plot(expt_time*1000,ss0_data,'ko')
        axs[i].set_xlim((0,25))
        axs[i].set_ylim((0.0,1.1))
        axs[i].get_xaxis().set_ticks([0,10,20])
        axs[i].title.set_text(res_type_sets)
        
        if i > 0:
            axs[i].get_yaxis().set_ticklabels([])
        i+=1
    fig.subplots_adjust(wspace=0,hspace=0)
    if export:
        fig.savefig('RMSD_best_points.pdf', format='pdf', dpi=800)
        
def plot_many_redor_point(dephasor_coordinates_list,expt_time,expt_data,coord_sets,export=False):
    '''
    Plots the type-REDOR curves for a many dephasor coordinates

    '''
    
    coup_1A = get_coupling_1A()
    
    n_types = len(expt_data.keys())
    fig,axs = plt.subplots(1,n_types,figsize=(12,6))
    
    sim_time = np.linspace(0,25,200)/1000
        
    i = 0
    for res_type_sets in expt_data.keys():
        
        entry = expt_data[res_type_sets]
        ss0_data = entry['SS0']
        err = entry['err']
        
        dephased_coordinates = coord_sets[res_type_sets]

        
        for dephasor_coordinates in dephasor_coordinates_list:
            
            distances = coords_to_distances(dephased_coordinates,dephasor_coordinates) # Convert to distances
            redor_sim = second_moment_redor(distances,coup_1A,sim_time)
            
            axs[i].plot(sim_time*1000,redor_sim,c='#1f77b4')
        
        #axs[i].plot(expt_time*1000,ss0_data,'ko')
        axs[i].errorbar(expt_time*1000,ss0_data,yerr=err,fmt='ko',capsize=4,ecolor='k')
        axs[i].set_xlim((0,25))
        axs[i].set_ylim((0.5,1.05))
        axs[i].get_xaxis().set_ticks([0,10,20])
        axs[i].title.set_text(res_type_sets)
        
        if i > 0:
            axs[i].get_yaxis().set_ticklabels([])
        i+=1
            
    fig.subplots_adjust(wspace=0,hspace=0)
    if export:
        fig.savefig('RMSD_best_points.pdf', format='pdf', dpi=800)
        
        
def plot_many_redor_points_plus(dephasor_coordinates_list,real_coord,expt_time,expt_data,coord_sets,export=False):
    '''
    Plots the type-REDOR curves for a many dephasor coordinates and a "real" coordinate

    '''
    coup_1A = get_coupling_1A()
    
    n_types = len(expt_data.keys())
    fig,axs = plt.subplots(1,n_types,figsize=(15,6))
    
    sim_time = np.linspace(0,25,200)/1000
        
    i = 0
    for res_type_sets in expt_data.keys():
        
        entry = expt_data[res_type_sets]
        ss0_data = entry['SS0']
        err = entry['err']
        
        dephased_coordinates = coord_sets[res_type_sets]

        
        for dephasor_coordinates in dephasor_coordinates_list:
            
            distances = coords_to_distances(dephased_coordinates,dephasor_coordinates) # Convert to distances
            redor_sim = second_moment_redor(distances,coup_1A,sim_time)
            
            axs[i].plot(sim_time*1000,redor_sim,c='#1f77b4')
        
        
        dists_real = coords_to_distances(dephased_coordinates,real_coord)
        real_redor = second_moment_redor(dists_real,coup_1A,sim_time)
        axs[i].plot(sim_time*1000,real_redor,'r')
        axs[i].errorbar(expt_time*1000,ss0_data,yerr=err,fmt='ko',capsize=4,ecolor='k')
        axs[i].set_xlim((0,25))
        axs[i].set_ylim((0.6,1.05))
        axs[i].get_xaxis().set_ticks([0,10,20])
        axs[i].title.set_text(res_type_sets)
        
        if i > 0:
            axs[i].get_yaxis().set_ticklabels([])
        i+=1
            
    fig.subplots_adjust(wspace=0,hspace=0)
    if export:
        fig.savefig('many_plus_1.pdf', format='pdf', dpi=800)
        
        
def get_coord_sets(expt_data, struct):
    '''
        Given an experimental data package and a Bio.PDB struct object, extracts the needed atoms.
        Returns them as a dictionary. 
    '''
    coordinate_sets = {}
    for atom_type_key in expt_data.keys():
        
        entry = expt_data[atom_type_key] # The entry for this set
        res_types = entry['res_types'] # A string like 'A' or 'DL'
        atoms = entry['atoms'] # A list like ['CA'] or ['CG1', 'CG2']
        
        all_coords = []
        n_coords = 0
        for res in res_types:
            # For each residue type group (e.q. 'I' or 'DL')
            let3 = structure_tools.convert_AA(res)
            coords = structure_tools.get_res_atom_coordinates(struct,let3,atoms) # Get to coordinates for those atoms (I-CA, V-CG1+CG2, etc)
            all_coords.append(coords)
            n_coords += len(coords)
            
        all_coords = np.vstack(all_coords)
        coordinate_sets[atom_type_key] = all_coords
        
    return coordinate_sets


if __name__ == '__main__':
    

    # Initialize (for all runs):
    T_0 = 1000
    T_min = 0.1
    alpha = 1.01
    coup_1A = get_coupling_1A() # Get the 1 Angstrom C-F coupling, including scaling
    T_scale = 10000
    
    # ## For GB1 Analysis, uncomment ##
    N_runs = 500
    struct = structure_tools.read_pdb('2LGI','2LGI.pdb') # Get the structure
    bounding_box = structure_tools.get_bounding_box(struct,margin=5) # Construct the bounding box
    original_dephasor_coordinates = np.array([[0,0,0]])
    (expt_time,expt_data) = get_experimental_data_GB1() # GB1 experiment data
    # In GB1, proper coordinate of the 19F: [4.102, 14.6, 18.067]
    w_coord_real = np.array([[4.102, 14.6, 18.067]])
    coordinate_sets = get_coord_sets(expt_data, struct) # Extract the atomic coordinates
    rmsd_real = get_point_rmsd(w_coord_real, coordinate_sets, expt_time, expt_data, coup_1A)
    fileprefix = 'GB1rev_'
    
    
    # ## For EmrE Analysis, uncomment ##
    #N_runs = 5000
    # struct = structure_tools.read_pdb('7JK8','7JK8_noloops.pdb',all_struct=False) # Get the structure
    # bounding_box = structure_tools.get_bounding_box(struct,margin=5) # Construct the bounding box
    # original_dephasor_coordinates = generate_tetrahedron(6.0)
    # (expt_time,expt_data) = get_experimental_data_EmrE() # GB1 experiment data
    # emre_actual_center = np.array([51.97,91.467,35.882])
    # emre_actual_rotation = R.from_euler('xyz',[ -76.87315582,-204.80254157 ,372.82392447])
    # emre_actual_coords_tet = rotate_translate(original_dephasor_coordinates,emre_actual_center,emre_actual_rotation)
    # emre_actual_coords = np.array([[47.715, 92.599, 39.985],
    #                                [56.682, 88.499, 38.104],
    #                                [48.812, 88.189, 31.988],
    #                                [53.51,  96.446, 32.944]])
    # coordinate_sets = get_coord_sets(expt_data, struct) # Extract the atomic coordinates
    # rmsd_real = get_point_rmsd(emre_actual_coords,coordinate_sets,expt_time,expt_data,coup_1A)
    # rmsd_best_tet = get_point_rmsd(emre_actual_coords_tet,coordinate_sets,expt_time,expt_data,coup_1A)
    # fileprefix = 'EmrE_'
    
    
    end_rmsds = np.array([])
    end_coords = []
    end_centers = []
    end_f_plus_centers = [] # For collision detection
    for run in range(N_runs):
        # BEGIN SIMULATED ANNEALING
        print(run)
         
        drug_center = get_random_coordinate(bounding_box) # Get a random starting point for the drug
        drug_rotation = get_initial_angles() # And a random set of euler angles
    
        dephasor_coordinates = rotate_translate(original_dephasor_coordinates,drug_center,drug_rotation)
        rmsd = get_point_rmsd(dephasor_coordinates,coordinate_sets,expt_time,expt_data,coup_1A) # Get the RMSD
        T = T_0 # Initialize the Temp
        
        all_rmsd = np.array([]) # Intialize array to store RMSD values
        while T > T_min:
            new_center = get_new_center(drug_center,10,bounding_box)
            new_rotation = get_new_angles(drug_rotation,30)
            new_dephasor_coordinates = rotate_translate(original_dephasor_coordinates,new_center,new_rotation)
            
            new_rmsd = get_point_rmsd(new_dephasor_coordinates,coordinate_sets,expt_time,expt_data,coup_1A)
            
            deltaE = (new_rmsd - rmsd)*T_scale # Get the energy difference (scaling here matches the temperature range)
            
            if acceptanceFunction(T,deltaE):
                drug_center = new_center
                drug_rotation = new_rotation
                dephasor_coordinates = new_dephasor_coordinates
                rmsd = new_rmsd
            T = T/alpha
            
            all_rmsd = np.append(all_rmsd,rmsd)
            
        #plt.plot(all_rmsd)
        
        end_rmsds = np.append(end_rmsds,rmsd)
        end_coords.append(dephasor_coordinates)
        end_centers.append(drug_center)
        
        end_f_plus_centers.append(np.vstack([dephasor_coordinates,drug_center]))
    
    
    end_coords = np.array(end_coords)
    end_centers = np.array(end_centers)
    print_pymol_commands(end_centers)
    plot_many_redor_point(end_coords, expt_time, expt_data, coordinate_sets)
    
    
    
    
    write_pymol_to_file(end_centers,fileprefix+'end_centers.txt')
    write_pymol_to_file(end_coords,fileprefix+'end_f_atoms.txt')
    write_pymol_to_file_rmsds(end_centers, end_rmsds, fileprefix+'end_centers_RMSD_color.txt')
    
    # Collision checking (for EmrE)
    # collisions_nohit_adv = structure_tools.check_collisions_emre(end_centers,end_coords,struct,center_margin = 2, corner_margin = 0.5)
    # end_coords_nohit_adv = end_coords[collisions_nohit_adv,:]
    # end_centers_nohit_adv = end_centers[collisions_nohit_adv,:]
    # write_pymol_to_file(end_centers_nohit_adv,fileprefix+'end_centers_nohit_adv.txt')
    # write_pymol_to_file(end_coords_nohit_adv,fileprefix+'end_f_atoms_nohit_adv.txt')
    
    #plot_many_redor_points_plus(end_coords_nohit_adv, emre_actual_coords_tet,expt_time, expt_data, coordinate_sets,export=True)
    plot_many_redor_points_plus(end_coords,w_coord_real,expt_time, expt_data, coordinate_sets,export=False)