#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Structure tools for Assignment-Free REDOR Analysis

This file contains helper functions for the analysis.

Written by Noah H. Somberg and Mei Hong

Correspondence to: meihong@mit.edu
"""
import numpy as np
import Bio
import Bio.PDB
import warnings

def check_AA(key,n=1):
    '''
    Checks if the input is a valid amino acid

    Parameters
    ----------
    key : str
        The AA type, i.e. 'G' for n=1 or 'GLY' for n=0
    n : str, optional
        n = 1 for 1-letter codes, n = 3 for 3-letter codes. The default is 1.

    Returns
    -------
    bool
        is valid AA? true or false.

    '''
    
    AA_3to1 = {'CYS': 'C', 'ASP': 'D', 'SER': 'S', 'GLN': 'Q', 'LYS': 'K',
         'ILE': 'I', 'PRO': 'P', 'THR': 'T', 'PHE': 'F', 'ASN': 'N', 
         'GLY': 'G', 'HIS': 'H', 'LEU': 'L', 'ARG': 'R', 'TRP': 'W', 
         'ALA': 'A', 'VAL':'V', 'GLU': 'E', 'TYR': 'Y', 'MET': 'M'}
    AA_1to3 = dict(map(reversed, AA_3to1.items()))
    
    if n == 1:
        if key in AA_1to3.keys():
            valid = True
        else:
            valid = False
    elif n == 3:
        if key in AA_3to1.keys():
            valid = True
        else:
            valid = False
            
    if valid:
        return
    else:
        raise Exception('Input was not a valid AA for n = '+str(n))
        
def convert_AA(in_str):
    
    AA_3to1 = {'CYS': 'C', 'ASP': 'D', 'SER': 'S', 'GLN': 'Q', 'LYS': 'K',
         'ILE': 'I', 'PRO': 'P', 'THR': 'T', 'PHE': 'F', 'ASN': 'N', 
         'GLY': 'G', 'HIS': 'H', 'LEU': 'L', 'ARG': 'R', 'TRP': 'W', 
         'ALA': 'A', 'VAL':'V', 'GLU': 'E', 'TYR': 'Y', 'MET': 'M'}
    AA_1to3 = dict(map(reversed, AA_3to1.items()))
    
    try:
        output = AA_3to1[in_str]
    except:
        try:
            output = AA_1to3[in_str]
        except:
            raise Exception('No AA found to convert for: '+str(in_str))\
                
    return output
            
    
def read_pdb(pdbcode, pdbfilenm,all_struct=True,folder='./'):
    '''
    Reads in a PDB file as a Bio.PDB.Structure
    
    Parameters
    ----------
    pdbcode : str
        4-letter PDB code
    pdbfilenm : str
        PDB file name

    Returns
    -------
    struct : Bio.PDB.Structure.Structure
        PDB file as a structure object
    '''
    pdbparser = Bio.PDB.PDBParser(QUIET=True)   # suppress PDBConstructionWarning
    struct = pdbparser.get_structure(pdbcode, folder+pdbfilenm)
    
    if all_struct:
        return struct
    else:
        return struct[0]

def get_all_calphas(struct):
    '''
    Takes a Bio.PDB.Structure and returns a list of all the CA atoms
    
    Parameters
    ----------
    struct : Bio.PDB.Structure.Structure
        PDB file as a structure object.

    Returns
    -------
    calphas : list
        All the structure CA atoms
    '''
    calphas = [ atom for atom in struct.get_atoms() if atom.get_fullname() == " CA " ]
    return calphas


def get_all_carbons(struct):
    '''
    Takes a Bio.PDB.Structure and returns a list of all the carbon atoms
    
    Parameters
    ----------
    struct : Bio.PDB.Structure.Structure
        PDB file as a structure object.

    Returns
    -------
    carbons : list
        All the structure carbons
    '''
    carbons = [atom for atom in struct.get_atoms() if 'C' in atom.get_fullname()]
    return carbons

def get_ca_residues(struct,res_type):
    '''
    Takes a Bio.PDB.Structure and a single res type, returns a list of all the CAs of that res type
    
    Parameters
    ----------
    struct : BioPDB structure
    res_type : str
        A three letter residue type, e.g. 'GLY'

    Returns
    -------
    atoms : list
        A list of atom objects.

    '''
    
    check_AA(res_type,n=3) # Check the AA type was valid
    
    residues = [res for res in struct.get_residues() if res.get_resname() == res_type] # Get all the residues of that type
    
    atoms = [] # Extract the CAs
    for res in residues:
        for atom in res.get_atoms():
            if atom.get_fullname() == " CA ":
                atoms.append(atom)
    
    return atoms

def get_residue_atoms_by_names(struct,res_type,atom_names):
    '''
    Takes in a Bio.PDB.Structure, a single residue type e.g. 'GLY', and a list of atoms to extract e.g. ['CA','CB']

    Parameters
    ----------
    struct : BioPDB structure
    res_type : str
        A three letter residue type, e.g. 'GLY'.
    atom_names : list
        List of atoms to extract, e.g. ['CA'] or ['CG1','CG2']

    Returns
    -------
    atoms : list
        List of the requested atoms

    '''
    # Check the atom names
    if not type(atom_names) is list:
        raise TypeError('Atom names must be list! Can be single-element')
    check_AA(res_type,n=3) # Check the AA type was valid
    
    # Get all of the residues of that type
    residues = [res for res in struct.get_residues() if res.get_resname() == res_type]
    
    expected_len = len(residues) * len(atom_names)
    
     # Look at all of the atoms in each residue. If the atom name is in the list atom_names, then take it
    atoms = []
    for res in residues:
        for atom in res.get_atoms():
            
            if atom.get_fullname().strip() in atom_names:
                atoms.append(atom)
               
                
    if len(atoms) != expected_len:
        unexpected_len = "For res_type = "+res_type+" atom_names = "+str(atom_names)+" expected "+str(expected_len)+" atoms but found "+str(len(atoms))
        warnings.warn(unexpected_len) 
        
    return atoms

def atoms_to_coords(atoms):
    '''
    Converts a list of N atoms to an N x 3 np array of atomic coordinates

    Parameters
    ----------
    atoms : list
        List of Bio.PDB.Atoms, len N

    Returns
    -------
    coords : np.array (N x 3)
        coordinates as an np.array

    '''
    
    coords = np.zeros((len(atoms),3)) # Initialize the array
    
    atom_idx = 0
    for atom in atoms:
        vec = np.array(list(atom.get_vector()))
        coords[atom_idx,:] = vec
        atom_idx += 1
        
    return coords

def get_res_atom_coordinates(struct,res_type,atom_names):
    '''
    Gets the atomic coordinates from a strucutre for the given res_type (e.g. 'ALA') and atom list (e.g.['CA','CB'])

    Parameters
    ----------
    struct : BioPDB structure
    res_type : str
        A three letter residue type, e.g. 'GLY'.
    atom_names : list
        List of atoms to extract, e.g. ['CA'] or ['CG1','CG2']
        
    Returns
    -------
    coords : np.array (N x 3)
        coordinates as an np.array

    '''
    
    atoms = get_residue_atoms_by_names(struct,res_type,atom_names)
    coords = atoms_to_coords(atoms)
    return coords

def get_bounding_box(struct,margin=5):
    """
    Gets valid x,y,z ranges for the structure

    Parameters
    ----------
    struct : BioPDB structure
    margin : float, optional
        Extra padding on the structure. The default is 5.

    Returns
    -------
    valid_ranges : np.array 3x2
        first row is [x_min, x_max], second is [y_min,y_max] ...
    """
    
    c_atoms = get_all_carbons(struct) # Get all the CA atoms in the structure
    coords = atoms_to_coords(c_atoms) # Get the CA coordinates
    
    min_vals = coords.min(axis=0)-margin
    max_vals = coords.max(axis=0)+margin
    
    valid_ranges = np.vstack((min_vals,max_vals)).transpose()
    
    return valid_ranges

def check_collisions(coordinates_list,struct,margin=1):
    
    all_atoms = [atom for atom in struct.get_atoms()]
    all_atom_coords = atoms_to_coords(all_atoms)

        
    collides = []
    for coordinates in coordinates_list:
        delta = np.linalg.norm(all_atom_coords[:,np.newaxis] - coordinates,axis=2)
        collisions = delta[delta < margin]
        if len(collisions) > 0:
            collides.append(False)
        else:
            collides.append(True)
                
    return collides

def check_collisions_emre(center_coords,corner_coords,struct,center_margin=3,corner_margin=1):
    
    # Go into residues, then atoms to avoid getting the drug atoms
    atoms = []
    for res in struct.get_residues():
        for atom in res.get_atoms():
               atoms.append(atom)
    all_atom_coords = atoms_to_coords(atoms)
    
    collides = []
    
    for i in range(len(center_coords)):
        curr_center = center_coords[i,:]
        curr_corners = corner_coords[i]
        
        center_delta = np.linalg.norm(all_atom_coords[:,np.newaxis] - curr_center,axis=2)
        corner_delta = np.linalg.norm(all_atom_coords[:,np.newaxis] - curr_corners,axis=2)
        
        center_collisions = center_delta[center_delta < center_margin]
        corner_collisions = corner_delta[corner_delta < corner_margin]
        
        if len(center_collisions) == 0 and len(corner_collisions) == 0:
            collides.append(True)
        else:
            collides.append(False)
            
    return collides
    
    

if __name__ == '__main__':
    # Some examples:
    struct = read_pdb('2lGI','2LGI.pdb')
    struct_gly_ca = get_residue_atoms_by_names(struct,'GLY',['CA','CB']) # Will warn, as GLY has no CB
    struct_gly_ca = get_residue_atoms_by_names(struct,'GLY',['CA']) # Won't warn
    gly_ca = get_res_atom_coordinates(struct,'GLY',['CA'])
    
    
    coords = np.array([[[-0.257     , 27.07600021, 23.99399948],[0,0,0]],[[100    , 100, 100],[100,100,100]]])
    res = check_collisions(coords, struct)

        
    
    a = get_bounding_box(struct)